@@ -35,9 +35,10 @@ from gi.repository import OsmGpsMap
 
 from . import stats
 
-from .gpx import GPXTrace
-
-from .utils.timezone import LocalTimezone
+from gpxpy import parse
+from gpxpy.gpx import GPXException
+from dateutil import tz
+from colorsys import hsv_to_rgb
 
 import locale
 import gettext
@@ -63,75 +64,48 @@ ALPHA_SELECTED = 0.8
 LAZY_LOAD_AFTER_N_FILES = 3
 
 
-class _TrackManager(GObject.GObject):
+class MainWindow:
     NAME_IDX = 0
-    FILENAME_IDX = 1
-
-    __gsignals__ = {
-        'track-added': (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE, [object, object]),
-        'track-removed': (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE, [object, object]),
-    }
-
-    def __init__(self):
-        GObject.GObject.__init__(self)
-        # maps track_filename : (GPXTrace, [OsmGpsMapTrack])
-        self._tracks = {}
-        # name, filename
-        self.model = Gtk.ListStore(str, str)
+    GPX_IDX = 1
+    OSM_IDX = 2
 
     def get_other_tracks(self, trace):
         tracks = []
-        for _trace, _tracks in self._tracks.values():
-            if trace != _trace:
-                tracks += _tracks
+        for row in self.model:
+            for track in row.iterchildren():
+                if trace != track[self.GPX_IDX]:
+                    tracks += track[self.OSM_IDX]
         return tracks
 
-    def get_trace_from_model(self, _iter):
-        filename = self.model.get_value(_iter, self.FILENAME_IDX)
-        return self.get_trace(filename)
+    def add_track(self, parent, track, color):
+        gpstracks = []
+        for segment in track.segments:
 
-    def delete_trace_from_model(self, _iter):
-        self.emit("track-removed", *self._tracks[self.model.get_value(_iter, self.FILENAME_IDX)])
-        self.model.remove(_iter)
+            gpstrack = OsmGpsMap.MapTrack()
+            gpstrack.set_color(color)
+            gpstrack.props.alpha = 0.8
 
-    def get_trace(self, filename):
-        """ Returns (trace, [OsmGpsMapTrack]) """
-        return self._tracks[filename]
+            for point in segment.points:
+                gpstrack.add_point(OsmGpsMap.MapPoint.new_degrees(point.latitude, point.longitude))
 
-    def add_trace(self, trace):
-        filename = trace.get_full_path()
-        if filename not in self._tracks:
-            gpstracks = []
-            for track in trace.get_points():
-                for segment in track:
+            gpstracks.append(gpstrack)
+            self.map.track_add(gpstrack)
 
-                    gpstrack = OsmGpsMap.MapTrack()
-                    gpstrack.props.alpha = 0.8
-
-                    for rlat, rlon in segment:
-                        gpstrack.add_point(OsmGpsMap.MapPoint.new_radians(rlat, rlon))
-                    gpstracks.append(gpstrack)
-
-            self._tracks[filename] = (trace, gpstracks)
-            self.model.append((trace.get_display_name(), filename))
-            self.emit("track-added", trace, gpstracks)
-
-    def num_traces(self):
-        return len(self._tracks)
+        self.model.append(parent, [track.name, track, gpstracks])
 
     def get_all_traces(self):
-        return [t[0] for t in self._tracks.values()]
+        return [t[self.GPX_IDX] for f in self.model for t in f.iterchildren()]
 
-
-class MainWindow:
     def __init__(self, ui_dir, files):
-        self.local_tz = LocalTimezone()
         self.recent = Gtk.RecentManager.get_default()
 
         self.wTree = Gtk.Builder()
         self.wTree.set_translation_domain('gpxviewer')
         self.wTree.add_from_file("%sgpxviewer.ui" % ui_dir)
 
+        # track_name, gpx, [OsmGpsMapTrack]
+        self.model = Gtk.TreeStore(str, object, object)
+
         signals = {
             "on_windowMain_destroy": self.quit,
             "on_menuitemQuit_activate": self.quit,
@@ -208,10 +182,6 @@ class MainWindow:
 
         self.wTree.get_object('menuitemOpenBy').set_submenu(submenu_open_with)
 
-        self.trackManager = _TrackManager()
-        self.trackManager.connect("track-added", self.on_track_added)
-        self.trackManager.connect("track-removed", self.on_track_removed)
-
         self.wTree.get_object("menuitemHelp").connect("activate",
                                                       lambda *a: show_url("https://answers.launchpad.net/gpxviewer"))
         self.wTree.get_object("menuitemTranslate").connect("activate", lambda *a: show_url(
@@ -219,13 +189,13 @@ class MainWindow:
         self.wTree.get_object("menuitemReportProblem").connect("activate", lambda *a: show_url(
             "https://bugs.launchpad.net/gpxviewer/+filebug"))
 
-        self.tv = Gtk.TreeView(self.trackManager.model)
+        self.tv = Gtk.TreeView(self.model)
         self.tv.get_selection().connect("changed", self.on_selection_changed)
         self.tv.append_column(
             Gtk.TreeViewColumn(
                 "Track Name",
                 Gtk.CellRendererText(),
-                text=self.trackManager.NAME_IDX
+                text=self.NAME_IDX
             )
         )
         self.wTree.get_object("scrolledwindow1").add(self.tv)
@@ -259,13 +229,11 @@ class MainWindow:
             i = 0
             for filename in files:
                 self.loadingFiles = i
-                trace = self.load_gpx(filename)
+                self.load_gpx(filename)
                 if i < LAZY_LOAD_AFTER_N_FILES:
                     i += 1
                 else:
-                    # select the last loaded trace
                     self.loadingFiles = 0
-                    self.select_trace(trace)
                     break
         else:
             self.loadingFiles = len(files)
@@ -299,22 +267,14 @@ class MainWindow:
         if not _iter:
             return
 
-        trace, tracks = self.trackManager.get_trace_from_model(_iter)
-        self.select_trace(trace)
+        trace = self.model.get_value(_iter, self.GPX_IDX)
+        tracks = self.model.get_value(_iter, self.OSM_IDX)
+        self.select_trace(self.model[_iter])
 
         # highlight current track
         self.select_tracks(tracks, ALPHA_SELECTED)
         # dim other tracks
-        self.select_tracks(self.trackManager.get_other_tracks(trace), ALPHA_UNSELECTED)
-
-    def on_track_added(self, tm, trace, tracks):
-        for t in tracks:
-            self.map.track_add(t)
-        self.select_trace(trace)
-
-    def on_track_removed(self, tm, trace, tracks):
-        for t in tracks:
-            self.map.track_remove(t)
+        self.select_tracks(self.get_other_tracks(trace), ALPHA_UNSELECTED)
 
     def update_tiles_queued(self, map_, paramspec):
         if self.map.props.tiles_queued > 0:
@@ -331,7 +291,7 @@ class MainWindow:
     def show_statistics(self, item):
         ws = stats.WeekStats()
         ss = stats.AvgSpeedStats()
-        for t in self.trackManager.get_all_traces():
+        for t in self.get_all_traces():
             ws.addTrace(t)
             ss.addTrace(t)
 
@@ -349,46 +309,67 @@ class MainWindow:
         dialog.show_all()
 
     def select_tracks(self, tracks, alpha):
+        if not tracks:
+            return
         for t in tracks:
             t.props.alpha = alpha
 
-    def select_trace(self, trace):
-        if self.loadingFiles:
+    def select_trace(self, row):
+        if not row[self.GPX_IDX]:
+            self.set_distance_label()
+            self.set_maximum_speed_label()
+            self.set_average_speed_label()
+            self.set_duration_label()
+            self.set_logging_date_label()
+            self.set_logging_time_label()
+
+            self.currentFilename = row[self.NAME_IDX]
+            self.mainWindow.set_title(_("GPX Viewer - %s") % row[self.NAME_IDX])
             return
 
         self.zoom = 12
-        distance = trace.get_distance()
-        maximum_speed = trace.get_maximum_speed()
-        average_speed = trace.get_average_speed()
-        duration = trace.get_duration()
-        clat, clon = trace.get_centre()
-        gpxfrom = trace.get_gpxfrom().astimezone(self.local_tz)
-        gpxto = trace.get_gpxto().astimezone(self.local_tz)
+        distance = row[self.GPX_IDX].get_moving_data().moving_distance
+        maximum_speed = row[self.GPX_IDX].get_moving_data().max_speed
+        average_speed = stats.get_average_speed(row[self.GPX_IDX])
+        duration = row[self.GPX_IDX].get_moving_data().moving_time
+        clat = row[self.GPX_IDX].get_center().latitude
+        clon = row[self.GPX_IDX].get_center().longitude
+        gpxfrom = row[self.GPX_IDX].segments[0].points[0].time.astimezone(tz.tzlocal())
+        gpxto = row[self.GPX_IDX].segments[-1].points[-1].time.astimezone(tz.tzlocal())
 
         self.set_distance_label(round(distance / 1000, 2))
         self.set_maximum_speed_label(maximum_speed)
         self.set_average_speed_label(average_speed)
-        self.set_duration_label(int(duration / 60), duration - (int(duration / 60) * 60))
+        hours, remain = divmod(duration, 3600)
+        minutes, seconds = divmod(remain, 60)
+        self.set_duration_label(hours, minutes, seconds)
         self.set_logging_date_label(gpxfrom.strftime("%x"))
         self.set_logging_time_label(gpxfrom.strftime("%X"), gpxto.strftime("%X"))
 
-        self.currentFilename = trace.get_filename()
-        self.mainWindow.set_title(_("GPX Viewer - %s") % trace.get_filename())
+        self.currentFilename = row.get_parent()[self.NAME_IDX]
+        self.mainWindow.set_title(_("GPX Viewer - %s") % row[self.GPX_IDX].name)
 
         if self.autoCenter:
             self.set_centre(clat, clon)
 
     def load_gpx(self, filename):
         try:
-            trace = GPXTrace(filename)
-            self.trackManager.add_trace(trace)
-            if self.trackManager.num_traces() > 1:
-                self.show_track_selector()
-            return trace
-        except Exception as e:
+            tracks = parse(open(filename)).tracks
+        except GPXException:
             self.show_gpx_error()
             return None
 
+        parent = self.model.append(None, [filename, None, None])
+        for i, track in enumerate(tracks):
+            color = Gdk.RGBA(*hsv_to_rgb((i / len(tracks) + 1 / 3) % 1.0, 1.0, 1.0))
+            self.add_track(parent, track, color)
+        if len(self.model) > 1 or len(tracks) > 1:
+            self.wTree.get_object("checkmenuitemShowSidebar").set_active(True)
+            self.show_track_selector()
+        else:
+            self.select_trace(next(self.model[0].iterchildren()))
+        return track
+
     def open_gpx(self, *args):
         filechooser = Gtk.FileChooserDialog(title=_("Choose a GPX file to Load"), action=Gtk.FileChooserAction.OPEN,
                                             parent=self.mainWindow)
@@ -431,18 +412,21 @@ class MainWindow:
     def set_centre(self, lat, lon):
         self.map.set_center_and_zoom(lat, lon, self.zoom)
 
-    def set_distance_label(self, distance="--"):
-        self.wTree.get_object("labelDistance").set_markup(_("<b>Distance:</b> %.2f km") % distance)
+    def set_distance_label(self, distance=None):
+        distance = '%.2f' % distance if distance else '--'
+        self.wTree.get_object("labelDistance").set_markup(_("<b>Distance:</b> %s km") % distance)
 
-    def set_average_speed_label(self, average_speed="--"):
-        self.wTree.get_object("labelAverageSpeed").set_markup(_("<b>Average Speed:</b> %.2f m/s") % average_speed)
+    def set_average_speed_label(self, average_speed=None):
+        average_speed = '%.2f' % average_speed if average_speed else '--'
+        self.wTree.get_object("labelAverageSpeed").set_markup(_("<b>Average Speed:</b> %s m/s") % average_speed)
 
-    def set_maximum_speed_label(self, maximum_speed="--"):
-        self.wTree.get_object("labelMaximumSpeed").set_markup(_("<b>Maximum Speed:</b> %.2f m/s") % maximum_speed)
+    def set_maximum_speed_label(self, maximum_speed=None):
+        maximum_speed = '%.2f' % maximum_speed if maximum_speed else '--'
+        self.wTree.get_object("labelMaximumSpeed").set_markup(_("<b>Maximum Speed:</b> %s m/s") % maximum_speed)
 
-    def set_duration_label(self, minutes="--", seconds="--"):
+    def set_duration_label(self, hours="--", minutes="--", seconds="--"):
         self.wTree.get_object("labelDuration").set_markup(
-            _("<b>Duration:</b> %(minutes)s minutes, %(seconds)s seconds") % {"minutes": minutes, "seconds": seconds})
+            _("<b>Duration:</b> %(hours)s hours, %(minutes)s minutes, %(seconds)s seconds") % {"hours": hours, "minutes": minutes, "seconds": seconds})
 
     def set_logging_date_label(self, gpxdate="--"):
         self.wTree.get_object("labelLoggingDate").set_markup(_("<b>Logging Date:</b> %s") % gpxdate)
@@ -457,17 +441,27 @@ class MainWindow:
     def button_track_add_clicked(self, *args):
         self.open_gpx()
 
+    def remove_track(self, tracks):
+        for t in tracks:
+            self.map.track_remove(t)
+
     def button_track_delete_clicked(self, *args):
         model, _iter = self.tv.get_selection().get_selected()
-        if _iter:
-            self.trackManager.delete_trace_from_model(_iter)
+        if not _iter:
+            return
+        if self.model.get_value(_iter, self.OSM_IDX):
+            self.remove_track(self.model.get_value(_iter, self.OSM_IDX))
+        else:
+            for child in self.model[_iter].iterchildren():
+                self.remove_track(child[self.OSM_IDX])
+        self.model.remove(_iter)
 
     def button_track_properties_clicked(self, *args):
         model, _iter = self.tv.get_selection().get_selected()
         if _iter:
-            trace, OsmGpsMapTracks = self.trackManager.get_trace_from_model(_iter)
+            OsmGpsMapTracks = self.model.get_value(_iter, self.OSM_IDX)
             colorseldlg = Gtk.ColorSelectionDialog("Select track color")
-            colorseldlg.get_color_selection().set_current_color(OsmGpsMapTracks[0].props.color)
+            colorseldlg.get_color_selection().set_current_color(OsmGpsMapTracks[0].props.color.to_color())
             result = colorseldlg.run()
             if result == Gtk.ResponseType.OK:
                 color = colorseldlg.get_color_selection().get_current_rgba()
