@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 #
 #  main.py - Launcher for GPX Viewer
 #
@@ -26,10 +26,11 @@ parent_dir = os.path.dirname(os.path.abspath(__file__))
 source_dir = os.path.join(parent_dir, "gpxviewer")
 sys.path.append(source_dir)
 
-import gtk.gdk
+import gi
+gi.require_version('Gdk', '3.0')
+
+from gi.repository import Gdk
 from gpxviewer.ui import MainWindow
- 
-gtk.gdk.threads_init()
 
 if len(sys.argv) > 2:
 	files = sys.argv[1:]
